@@ -19,12 +19,20 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+// Maintainers, keep in mind that octal literals are not allowed
+// in strict mode. Use the decimal value and add a comment with
+// the octal value. Example:
+//
+//   var mode = 438; /* mode=0666 */
+
 var util = require('util');
+var pathModule = require('path');
 
 var binding = process.binding('fs');
 var constants = process.binding('constants');
 var fs = exports;
 var Stream = require('stream').Stream;
+var EventEmitter = require('events').EventEmitter;
 
 var kMinPoolSpace = 128;
 var kPoolSize = 40 * 1024;
@@ -109,7 +117,7 @@ fs.readFile = function(path, encoding_) {
 };
 
 fs.readFileSync = function(path, encoding) {
-  var fd = fs.openSync(path, constants.O_RDONLY, 0666);
+  var fd = fs.openSync(path, constants.O_RDONLY, 438 /*=0666*/);
   var buffer = new Buffer(4048);
   var buffers = [];
   var nread = 0;
@@ -193,7 +201,7 @@ fs.closeSync = function(fd) {
 };
 
 function modeNum(m, def) {
-  switch(typeof m) {
+  switch (typeof m) {
     case 'number': return m;
     case 'string': return parseInt(m, 8);
     default:
@@ -211,14 +219,15 @@ fs.open = function(path, flags, mode, callback) {
     callback = noop;
   }
 
-  mode = modeNum(mode, '0666');
+  mode = modeNum(mode, 438 /*=0666*/);
 
-  binding.open(path, stringToFlags(flags), mode, callback);
+  binding.open(pathModule._makeLong(path), stringToFlags(flags), mode,
+    callback);
 };
 
 fs.openSync = function(path, flags, mode) {
-  mode = modeNum(mode, '0666');
-  return binding.open(path, stringToFlags(flags), mode);
+  mode = modeNum(mode, 438 /*=0666*/);
+  return binding.open(pathModule._makeLong(path), stringToFlags(flags), mode);
 };
 
 fs.read = function(fd, buffer, offset, length, position, callback) {
@@ -313,11 +322,13 @@ fs.writeSync = function(fd, buffer, offset, length, position) {
 };
 
 fs.rename = function(oldPath, newPath, callback) {
-  binding.rename(oldPath, newPath, callback || noop);
+  binding.rename(pathModule._makeLong(oldPath), pathModule._makeLong(newPath),
+    callback || noop);
 };
 
 fs.renameSync = function(oldPath, newPath) {
-  return binding.rename(oldPath, newPath);
+  return binding.rename(pathModule._makeLong(oldPath),
+    pathModule._makeLong(newPath));
 };
 
 fs.truncate = function(fd, len, callback) {
@@ -329,11 +340,11 @@ fs.truncateSync = function(fd, len) {
 };
 
 fs.rmdir = function(path, callback) {
-  binding.rmdir(path, callback || noop);
+  binding.rmdir(pathModule._makeLong(path), callback || noop);
 };
 
 fs.rmdirSync = function(path) {
-  return binding.rmdir(path);
+  return binding.rmdir(pathModule._makeLong(path));
 };
 
 fs.fdatasync = function(fd, callback) {
@@ -353,11 +364,14 @@ fs.fsyncSync = function(fd) {
 };
 
 fs.mkdir = function(path, mode, callback) {
-  binding.mkdir(path, modeNum(mode), callback || noop);
+  if (typeof mode === 'function') callback = mode;
+  binding.mkdir(pathModule._makeLong(path), modeNum(mode, 511 /*=0777*/),
+    callback || noop);
 };
 
 fs.mkdirSync = function(path, mode) {
-  return binding.mkdir(path, modeNum(mode));
+  return binding.mkdir(pathModule._makeLong(path),
+    modeNum(mode, 511 /*=0777*/));
 };
 
 fs.sendfile = function(outFd, inFd, inOffset, length, callback) {
@@ -369,11 +383,11 @@ fs.sendfileSync = function(outFd, inFd, inOffset, length) {
 };
 
 fs.readdir = function(path, callback) {
-  binding.readdir(path, callback || noop);
+  binding.readdir(pathModule._makeLong(path), callback || noop);
 };
 
 fs.readdirSync = function(path) {
-  return binding.readdir(path);
+  return binding.readdir(pathModule._makeLong(path));
 };
 
 fs.fstat = function(fd, callback) {
@@ -381,11 +395,11 @@ fs.fstat = function(fd, callback) {
 };
 
 fs.lstat = function(path, callback) {
-  binding.lstat(path, callback || noop);
+  binding.lstat(pathModule._makeLong(path), callback || noop);
 };
 
 fs.stat = function(path, callback) {
-  binding.stat(path, callback || noop);
+  binding.stat(pathModule._makeLong(path), callback || noop);
 };
 
 fs.fstatSync = function(fd) {
@@ -393,59 +407,179 @@ fs.fstatSync = function(fd) {
 };
 
 fs.lstatSync = function(path) {
-  return binding.lstat(path);
+  return binding.lstat(pathModule._makeLong(path));
 };
 
 fs.statSync = function(path) {
-  return binding.stat(path);
+  return binding.stat(pathModule._makeLong(path));
 };
 
 fs.readlink = function(path, callback) {
-  binding.readlink(path, callback || noop);
+  binding.readlink(pathModule._makeLong(path), callback || noop);
 };
 
 fs.readlinkSync = function(path) {
-  return binding.readlink(path);
+  return binding.readlink(pathModule._makeLong(path));
 };
 
-fs.symlink = function(destination, path, callback) {
-  binding.symlink(destination, path, callback || noop);
+fs.symlink = function(destination, path, type_, callback) {
+  var type = (typeof(type_) == 'string' ? type_ : null);
+  var callback_ = arguments[arguments.length - 1];
+  callback = (typeof(callback_) == 'function' ? callback_ : null);
+  binding.symlink(pathModule._makeLong(destination),
+    pathModule._makeLong(path), type, callback);
 };
 
-fs.symlinkSync = function(destination, path) {
-  return binding.symlink(destination, path);
+fs.symlinkSync = function(destination, path, type) {
+  return binding.symlink(pathModule._makeLong(destination),
+    pathModule._makeLong(path), type);
 };
 
 fs.link = function(srcpath, dstpath, callback) {
-  binding.link(srcpath, dstpath, callback || noop);
+  binding.link(pathModule._makeLong(srcpath), pathModule._makeLong(dstpath),
+    callback || noop);
 };
 
 fs.linkSync = function(srcpath, dstpath) {
-  return binding.link(srcpath, dstpath);
+  return binding.link(pathModule._makeLong(srcpath),
+    pathModule._makeLong(dstpath));
 };
 
 fs.unlink = function(path, callback) {
-  binding.unlink(path, callback || noop);
+  binding.unlink(pathModule._makeLong(path), callback || noop);
 };
 
 fs.unlinkSync = function(path) {
-  return binding.unlink(path);
+  return binding.unlink(pathModule._makeLong(path));
 };
 
+fs.fchmod = function(fd, mode, callback) {
+  binding.fchmod(fd, modeNum(mode), callback || noop);
+};
+
+fs.fchmodSync = function(fd, mode) {
+  return binding.fchmod(fd, modeNum(mode));
+};
+
+if (constants.hasOwnProperty('O_SYMLINK')) {
+  fs.lchmod = function(path, mode, callback) {
+    callback = callback || noop;
+    fs.open(path, constants.O_WRONLY | constants.O_SYMLINK, function(err, fd) {
+      if (err) {
+        callback(err);
+        return;
+      }
+      // prefer to return the chmod error, if one occurs,
+      // but still try to close, and report closing errors if they occur.
+      fs.fchmod(fd, mode, function(err) {
+        fs.close(fd, function(err2) {
+          callback(err || err2);
+        });
+      });
+    });
+  };
+
+  fs.lchmodSync = function(path, mode) {
+    var fd = fs.openSync(path, constants.O_WRONLY | constants.O_SYMLINK);
+
+    // prefer to return the chmod error, if one occurs,
+    // but still try to close, and report closing errors if they occur.
+    var err, err2;
+    try {
+      var ret = fs.fchmodSync(fd, mode);
+    } catch (er) {
+      err = er;
+    }
+    try {
+      fs.closeSync(fd);
+    } catch (er) {
+      err2 = er;
+    }
+    if (err || err2) throw (err || err2);
+    return ret;
+  };
+}
+
+
 fs.chmod = function(path, mode, callback) {
-  binding.chmod(path, modeNum(mode), callback || noop);
+  binding.chmod(pathModule._makeLong(path), modeNum(mode), callback || noop);
 };
 
 fs.chmodSync = function(path, mode) {
-  return binding.chmod(path, modeNum(mode));
+  return binding.chmod(pathModule._makeLong(path), modeNum(mode));
+};
+
+if (constants.hasOwnProperty('O_SYMLINK')) {
+  fs.lchown = function(path, uid, gid, callback) {
+    callback = callback || noop;
+    fs.open(path, constants.O_WRONLY | constants.O_SYMLINK, function(err, fd) {
+      if (err) {
+        callback(err);
+        return;
+      }
+      fs.fchown(fd, uid, gid, callback);
+    });
+  };
+
+  fs.lchownSync = function(path, uid, gid) {
+    var fd = fs.openSync(path, constants.O_WRONLY | constants.O_SYMLINK);
+    return fs.fchownSync(fd, uid, gid);
+  };
+}
+
+fs.fchown = function(fd, uid, gid, callback) {
+  binding.fchown(fd, uid, gid, callback || noop);
+};
+
+fs.fchownSync = function(fd, uid, gid) {
+  return binding.fchown(fd, uid, gid);
 };
 
 fs.chown = function(path, uid, gid, callback) {
-  binding.chown(path, uid, gid, callback || noop);
+  binding.chown(pathModule._makeLong(path), uid, gid, callback || noop);
 };
 
 fs.chownSync = function(path, uid, gid) {
-  return binding.chown(path, uid, gid);
+  return binding.chown(pathModule._makeLong(path), uid, gid);
+};
+
+// converts Date or number to a fractional UNIX timestamp
+function toUnixTimestamp(time) {
+  if (typeof time == 'number') {
+    return time;
+  }
+  if (time instanceof Date) {
+    // convert to 123.456 UNIX timestamp
+    return time.getTime() / 1000;
+  }
+  throw new Error('Cannot parse time: ' + time);
+}
+
+// exported for unit tests, not for public consumption
+fs._toUnixTimestamp = toUnixTimestamp;
+
+fs.utimes = function(path, atime, mtime, callback) {
+  atime = toUnixTimestamp(atime);
+  mtime = toUnixTimestamp(mtime);
+  binding.utimes(pathModule._makeLong(path), atime, mtime, callback || noop);
+};
+
+fs.utimesSync = function(path, atime, mtime) {
+  atime = toUnixTimestamp(atime);
+  mtime = toUnixTimestamp(mtime);
+  binding.utimes(pathModule._makeLong(path), atime, mtime);
+};
+
+fs.futimes = function(fd, atime, mtime, callback) {
+  atime = toUnixTimestamp(atime);
+  mtime = toUnixTimestamp(mtime);
+  binding.futimes(fd, atime, mtime, callback || noop);
+};
+
+fs.futimesSync = function(fd, atime, mtime) {
+  atime = toUnixTimestamp(atime);
+  mtime = toUnixTimestamp(mtime);
+  binding.futimes(fd, atime, mtime);
 };
 
 function writeAll(fd, buffer, offset, length, callback) {
@@ -468,12 +602,13 @@ function writeAll(fd, buffer, offset, length, callback) {
 fs.writeFile = function(path, data, encoding_, callback) {
   var encoding = (typeof(encoding_) == 'string' ? encoding_ : 'utf8');
   var callback_ = arguments[arguments.length - 1];
-  var callback = (typeof(callback_) == 'function' ? callback_ : null);
-  fs.open(path, 'w', 0666, function(openErr, fd) {
+  callback = (typeof(callback_) == 'function' ? callback_ : null);
+  fs.open(path, 'w', 438 /*=0666*/, function(openErr, fd) {
     if (openErr) {
       if (callback) callback(openErr);
     } else {
-      var buffer = Buffer.isBuffer(data) ? data : new Buffer(data, encoding);
+      var buffer = Buffer.isBuffer(data) ? data : new Buffer('' + data,
+          encoding);
       writeAll(fd, buffer, 0, buffer.length, callback);
     }
   });
@@ -482,7 +617,7 @@ fs.writeFile = function(path, data, encoding_, callback) {
 fs.writeFileSync = function(path, data, encoding) {
   var fd = fs.openSync(path, 'w');
   if (!Buffer.isBuffer(data)) {
-    data = new Buffer(data, encoding || 'utf8');
+    data = new Buffer('' + data, encoding || 'utf8');
   }
   var written = 0;
   var length = data.length;
@@ -493,11 +628,112 @@ fs.writeFileSync = function(path, data, encoding) {
   fs.closeSync(fd);
 };
 
+
+function errnoException(errorno, syscall) {
+  // TODO make this more compatible with ErrnoException from src/node.cc
+  // Once all of Node is using this function the ErrnoException from
+  // src/node.cc should be removed.
+  var e = new Error(syscall + ' ' + errorno);
+  e.errno = e.code = errorno;
+  e.syscall = syscall;
+  return e;
+}
+
+
+function FSWatcher() {
+  var self = this;
+  var FSEvent = process.binding('fs_event_wrap').FSEvent;
+  this._handle = new FSEvent();
+
+  this._handle.onchange = function(status, event, filename) {
+    if (status) {
+      self.emit('error', errnoException(errno, 'watch'));
+    } else {
+      self.emit('change', event, filename);
+    }
+  };
+}
+util.inherits(FSWatcher, EventEmitter);
+
+FSWatcher.prototype.start = function(filename, persistent) {
+  var r = this._handle.start(pathModule._makeLong(filename), persistent);
+
+  if (r) {
+    this._handle.close();
+    throw errnoException(errno, 'watch');
+  }
+};
+
+FSWatcher.prototype.close = function() {
+  this._handle.close();
+};
+
+fs.watch = function(filename) {
+  var watcher;
+  var options;
+  var listener;
+
+  if ('object' == typeof arguments[1]) {
+    options = arguments[1];
+    listener = arguments[2];
+  } else {
+    options = {};
+    listener = arguments[1];
+  }
+
+  if (!listener) {
+    throw new Error('watch requires a listener function');
+  }
+
+  if (options.persistent === undefined) options.persistent = true;
+
+  watcher = new FSWatcher();
+  watcher.start(filename, options.persistent);
+
+  watcher.addListener('change', listener);
+  return watcher;
+};
+
+
 // Stat Change Watchers
 
+function StatWatcher() {
+  var self = this;
+  this._handle = new binding.StatWatcher();
+
+  this._handle.onchange = function(current, previous) {
+    self.emit('change', current, previous);
+  };
+
+  this._handle.onstop = function() {
+    self.emit('stop');
+  };
+}
+util.inherits(StatWatcher, EventEmitter);
+
+
+StatWatcher.prototype.start = function(filename, persistent, interval) {
+  this._handle.start(pathModule._makeLong(filename), persistent, interval);
+};
+
+
+StatWatcher.prototype.stop = function() {
+  this._handle.stop();
+};
+
+
 var statWatchers = {};
+function inStatWatchers(filename) {
+  return Object.prototype.hasOwnProperty.call(statWatchers, filename) &&
+      statWatchers[filename];
+}
+
 
 fs.watchFile = function(filename) {
+  if (isWindows) {
+    throw new Error('use fs.watch api instead');
+  }
+
   var stat;
   var options;
   var listener;
@@ -510,14 +746,17 @@ fs.watchFile = function(filename) {
     listener = arguments[1];
   }
 
+  if (!listener) {
+    throw new Error('watchFile requires a listener function');
+  }
+
   if (options.persistent === undefined) options.persistent = true;
   if (options.interval === undefined) options.interval = 0;
 
-  if (statWatchers[filename]) {
+  if (inStatWatchers(filename)) {
     stat = statWatchers[filename];
   } else {
-    statWatchers[filename] = new binding.StatWatcher();
-    stat = statWatchers[filename];
+    stat = statWatchers[filename] = new StatWatcher();
     stat.start(filename, options.persistent, options.interval);
   }
   stat.addListener('change', listener);
@@ -526,7 +765,7 @@ fs.watchFile = function(filename) {
 
 fs.unwatchFile = function(filename) {
   var stat;
-  if (statWatchers[filename]) {
+  if (inStatWatchers(filename)) {
     stat = statWatchers[filename];
     stat.stop();
     statWatchers[filename] = undefined;
@@ -537,8 +776,7 @@ fs.unwatchFile = function(filename) {
 // Not using realpath(2) because it's bad.
 // See: http://insanecoding.blogspot.com/2007/11/pathmax-simply-isnt.html
 
-var path = require('path'),
-    normalize = path.normalize,
+var normalize = pathModule.normalize,
     isWindows = process.platform === 'win32';
 
 if (isWindows) {
@@ -547,7 +785,7 @@ if (isWindows) {
 
   // windows version
   fs.realpathSync = function realpathSync(p, cache) {
-    var p = path.resolve(p);
+    p = pathModule.resolve(p);
     if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
       return cache[p];
     }
@@ -562,12 +800,12 @@ if (isWindows) {
       cb = cache;
       cache = null;
     }
-    var p = path.resolve(p);
+    p = pathModule.resolve(p);
     if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
       return cb(null, cache[p]);
     }
     fs.stat(p, function(err) {
-      if (err) cb(err);
+      if (err) return cb(err);
       if (cache) cache[p] = p;
       cb(null, p);
     });
@@ -583,7 +821,7 @@ if (isWindows) {
   // posix version
   fs.realpathSync = function realpathSync(p, cache) {
     // make p is absolute
-    p = path.resolve(p);
+    p = pathModule.resolve(p);
 
     if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
       return cache[p];
@@ -636,14 +874,14 @@ if (isWindows) {
         if (!seenLinks[id]) {
           fs.statSync(base);
           seenLinks[id] = fs.readlinkSync(base);
-          resolvedLink = path.resolve(previous, seenLinks[id]);
+          resolvedLink = pathModule.resolve(previous, seenLinks[id]);
           // track this, if given a cache.
           if (cache) cache[base] = resolvedLink;
         }
       }
 
       // resolve the link, then start over
-      p = path.resolve(resolvedLink, p.slice(pos));
+      p = pathModule.resolve(resolvedLink, p.slice(pos));
       pos = 0;
       previous = base = current = '';
     }
@@ -662,7 +900,7 @@ if (isWindows) {
     }
 
     // make p is absolute
-    p = path.resolve(p);
+    p = pathModule.resolve(p);
 
     if (cache && Object.prototype.hasOwnProperty.call(cache, p)) {
       return cb(null, cache[p]);
@@ -740,7 +978,7 @@ if (isWindows) {
     function gotTarget(err, target, base) {
       if (err) return cb(err);
 
-      var resolvedLink = path.resolve(previous, target);
+      var resolvedLink = pathModule.resolve(previous, target);
       if (cache) cache[base] = resolvedLink;
       gotResolvedLink(resolvedLink);
     }
@@ -748,7 +986,7 @@ if (isWindows) {
     function gotResolvedLink(resolvedLink) {
 
       // resolve the link, then start over
-      p = path.resolve(resolvedLink, p.slice(pos));
+      p = pathModule.resolve(resolvedLink, p.slice(pos));
       pos = 0;
       previous = base = current = '';
 
@@ -785,7 +1023,7 @@ var ReadStream = fs.ReadStream = function(path, options) {
   this.paused = false;
 
   this.flags = 'r';
-  this.mode = parseInt('0666', 8);
+  this.mode = 438; /*=0666*/
   this.bufferSize = 64 * 1024;
 
   options = options || {};
@@ -805,10 +1043,10 @@ var ReadStream = fs.ReadStream = function(path, options) {
     }
 
     if (this.start > this.end) {
-      this.emit('error', new Error('start must be <= end'));
-    } else {
-      this._firstRead = true;
+      throw new Error('start must be <= end');
     }
+
+    this.pos = this.start;
   }
 
   if (this.fd !== null) {
@@ -839,9 +1077,9 @@ ReadStream.prototype.setEncoding = function(encoding) {
 
 ReadStream.prototype._read = function() {
   var self = this;
-  if (!self.readable || self.paused || self.reading) return;
+  if (!this.readable || this.paused || this.reading) return;
 
-  self.reading = true;
+  this.reading = true;
 
   if (!pool || pool.length - pool.used < kMinPoolSpace) {
     // discard the old pool. Can't add to the free list because
@@ -850,16 +1088,11 @@ ReadStream.prototype._read = function() {
     allocNewPool();
   }
 
-  if (self.start !== undefined && self._firstRead) {
-    self.pos = self.start;
-    self._firstRead = false;
-  }
-
   // Grab another reference to the pool in the case that while we're in the
   // thread pool another read() finishes up the pool, and allocates a new
   // one.
   var thisPool = pool;
-  var toRead = Math.min(pool.length - pool.used, this.bufferSize);
+  var toRead = Math.min(pool.length - pool.used, ~~this.bufferSize);
   var start = pool.used;
 
   if (this.pos !== undefined) {
@@ -899,10 +1132,10 @@ ReadStream.prototype._read = function() {
     self._read();
   }
 
-  fs.read(self.fd, pool, pool.used, toRead, self.pos, afterRead);
+  fs.read(this.fd, pool, pool.used, toRead, this.pos, afterRead);
 
-  if (self.pos !== undefined) {
-    self.pos += toRead;
+  if (this.pos !== undefined) {
+    this.pos += toRead;
   }
   pool.used += toRead;
 };
@@ -979,7 +1212,8 @@ var WriteStream = fs.WriteStream = function(path, options) {
 
   this.flags = 'w';
   this.encoding = 'binary';
-  this.mode = parseInt('0666', 8);
+  this.mode = 438; /*=0666*/
+  this.bytesWritten = 0;
 
   options = options || {};
 
@@ -990,6 +1224,14 @@ var WriteStream = fs.WriteStream = function(path, options) {
     this[key] = options[key];
   }
 
+  if (this.start !== undefined) {
+    if (this.start < 0) {
+      throw new Error('start must be >= zero');
+    }
+
+    this.pos = this.start;
+  }
+
   this.busy = false;
   this._queue = [];
 
@@ -1008,7 +1250,7 @@ WriteStream.prototype.flush = function() {
 
   var args = this._queue.shift();
   if (!args) {
-    if (this.drainable) { self.emit('drain'); }
+    if (this.drainable) { this.emit('drain'); }
     return;
   }
 
@@ -1017,8 +1259,6 @@ WriteStream.prototype.flush = function() {
   var method = args.shift(),
       cb = args.pop();
 
-  var self = this;
-
   args.push(function(err) {
     self.busy = false;
 
@@ -1031,8 +1271,20 @@ WriteStream.prototype.flush = function() {
       return;
     }
 
-    // stop flushing after close
-    if (method === fs.close) {
+    if (method == fs.write) {
+      self.bytesWritten += arguments[1];
+      if (cb) {
+        // write callback
+        cb(null, arguments[1]);
+      }
+
+    } else if (method === fs.open) {
+      // save reference for file pointer
+      self.fd = arguments[1];
+      self.emit('open', self.fd);
+
+    } else if (method === fs.close) {
+      // stop flushing after close
       if (cb) {
         cb(null);
       }
@@ -1040,21 +1292,12 @@ WriteStream.prototype.flush = function() {
       return;
     }
 
-    // save reference for file pointer
-    if (method === fs.open) {
-      self.fd = arguments[1];
-      self.emit('open', self.fd);
-    } else if (cb) {
-      // write callback
-      cb(null, arguments[1]);
-    }
-
     self.flush();
   });
 
   // Inject the file pointer
   if (method !== fs.open) {
-    args.unshift(self.fd);
+    args.unshift(this.fd);
   }
 
   method.apply(this, args);
@@ -1062,7 +1305,8 @@ WriteStream.prototype.flush = function() {
 
 WriteStream.prototype.write = function(data) {
   if (!this.writable) {
-    throw new Error('stream not writable');
+    this.emit('error', new Error('stream not writable'));
+    return false;
   }
 
   this.drainable = true;
@@ -1072,14 +1316,17 @@ WriteStream.prototype.write = function(data) {
     cb = arguments[arguments.length - 1];
   }
 
-  if (Buffer.isBuffer(data)) {
-    this._queue.push([fs.write, data, 0, data.length, null, cb]);
-  } else {
+  if (!Buffer.isBuffer(data)) {
     var encoding = 'utf8';
     if (typeof(arguments[1]) == 'string') encoding = arguments[1];
-    this._queue.push([fs.write, data, undefined, encoding, cb]);
+    data = new Buffer('' + data, encoding);
   }
 
+  this._queue.push([fs.write, data, 0, data.length, this.pos, cb]);
+
+  if (this.pos !== undefined) {
+    this.pos += data.length;
+  }
 
   this.flush();
 
@@ -1127,3 +1374,64 @@ WriteStream.prototype.destroy = function(cb) {
 // There is no shutdown() for files.
 WriteStream.prototype.destroySoon = WriteStream.prototype.end;
 
+
+// SyncWriteStream is internal. DO NOT USE.
+// Temporary hack for process.stdout and process.stderr when piped to files.
+function SyncWriteStream(fd) {
+  this.fd = fd;
+  this.writable = true;
+  this.readable = false;
+};
+util.inherits(SyncWriteStream, Stream);
+
+
+// Export
+fs.SyncWriteStream = SyncWriteStream;
+
+
+SyncWriteStream.prototype.write = function(data, arg1, arg2) {
+  var encoding, cb;
+
+  // parse arguments
+  if (arg1) {
+    if (typeof arg1 === 'string') {
+      encoding = arg1;
+      cb = arg2;
+    } else if (typeof arg1 === 'function') {
+      cb = arg1;
+    } else {
+      throw new Error("bad arg");
+    }
+  }
+
+  // Change strings to buffers. SLOW
+  if (typeof data == 'string') {
+    data = new Buffer(data, encoding);
+  }
+
+  fs.writeSync(this.fd, data, 0, data.length);
+
+  if (cb) {
+    process.nextTick(cb);
+  }
+
+  return true;
+};
+
+
+SyncWriteStream.prototype.end = function(data, arg1, arg2) {
+  if (data) {
+    this.write(data, arg1, arg2);
+  }
+  this.destroy();
+};
+
+
+SyncWriteStream.prototype.destroy = function() {
+  fs.closeSync(this.fd);
+  this.fd = null;
+  this.emit('close');
+  return true;
+};
+
+SyncWriteStream.prototype.destroySoon = SyncWriteStream.prototype.destroy;
