@@ -20,7 +20,7 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 var NativeModule = require('native_module');
-var Script = process.binding('evals').Script;
+var Script = process.binding('evals').NodeScript;
 var runInThisContext = Script.runInThisContext;
 var runInNewContext = Script.runInNewContext;
 var assert = require('assert').ok;
@@ -52,7 +52,8 @@ Module._contextLoad = (+process.env['NODE_MODULE_CONTEXTS'] > 0);
 Module._cache = {};
 Module._pathCache = {};
 Module._extensions = {};
-Module._paths = [];
+var modulePaths = [];
+Module.globalPaths = [];
 
 Module.wrapper = NativeModule.wrapper;
 Module.wrap = NativeModule.wrap;
@@ -102,11 +103,18 @@ function readPackage(requestPath) {
   try {
     var jsonPath = path.resolve(requestPath, 'package.json');
     var json = fs.readFileSync(jsonPath, 'utf8');
-    var pkg = packageCache[requestPath] = JSON.parse(json);
-    return pkg;
-  } catch (e) {}
+  } catch (e) {
+    return false;
+  }
 
-  return false;
+  try {
+    var pkg = packageCache[requestPath] = JSON.parse(json);
+  } catch (e) {
+    e.path = jsonPath;
+    e.message = 'Error parsing ' + jsonPath + ': ' + e.message;
+    throw e;
+  }
+  return pkg;
 }
 
 function tryPackage(requestPath, exts) {
@@ -122,7 +130,7 @@ function tryPackage(requestPath, exts) {
 // In order to minimize unnecessary lstat() calls,
 // this cache is a list of known-real paths.
 // Set to an empty object to reset.
-Module._realpathCache = {}
+Module._realpathCache = {};
 
 // check if the file exists and is not a directory
 function tryFile(requestPath) {
@@ -208,7 +216,7 @@ Module._nodeModulePaths = function(from) {
   var paths = [];
   var parts = from.split(splitRe);
 
-  for (var tip = parts.length - 1; tip >= 0; tip --) {
+  for (var tip = parts.length - 1; tip >= 0; tip--) {
     // don't search in .../node_modules/node_modules
     if (parts[tip] === 'node_modules') continue;
     var dir = parts.slice(0, tip + 1).concat('node_modules').join(joiner);
@@ -216,7 +224,7 @@ Module._nodeModulePaths = function(from) {
   }
 
   return paths;
-}
+};
 
 
 Module._resolveLookupPaths = function(request, parent) {
@@ -226,7 +234,7 @@ Module._resolveLookupPaths = function(request, parent) {
 
   var start = request.substring(0, 2);
   if (start !== './' && start !== '..') {
-    var paths = Module._paths;
+    var paths = modulePaths;
     if (parent) {
       if (!parent.paths) parent.paths = [];
       paths = parent.paths.concat(paths);
@@ -238,7 +246,7 @@ Module._resolveLookupPaths = function(request, parent) {
   if (!parent || !parent.id || !parent.filename) {
     // make require('./path/to/foo') work - normally the path is taken
     // from realpath(__filename) but with eval there is no filename
-    var mainPaths = ['.'].concat(Module._paths);
+    var mainPaths = ['.'].concat(modulePaths);
     mainPaths = Module._nodeModulePaths('.').concat(mainPaths);
     return [request, mainPaths];
   }
@@ -345,6 +353,11 @@ Module.prototype.load = function(filename) {
 };
 
 
+Module.prototype.require = function(path) {
+  return Module._load(path, this);
+};
+
+
 // Returns exception if any
 Module.prototype._compile = function(content, filename) {
   var self = this;
@@ -352,20 +365,28 @@ Module.prototype._compile = function(content, filename) {
   content = content.replace(/^\#\!.*/, '');
 
   function require(path) {
-    return Module._load(path, self);
+    return self.require(path);
   }
 
   require.resolve = function(request) {
     return Module._resolveFilename(request, self)[1];
-  }
-  require.paths = Module._paths;
+  };
+
+  Object.defineProperty(require, 'paths', { get: function() {
+    throw new Error('require.paths is removed. Use ' +
+                    'node_modules folders, or the NODE_PATH ' +
+                    'environment variable instead.');
+  }});
+
   require.main = process.mainModule;
+
   // Enable support to add extra extension types
   require.extensions = Module._extensions;
   require.registerExtension = function() {
     throw new Error('require.registerExtension() removed. Use ' +
                     'require.extensions instead.');
-  }
+  };
+
   require.cache = Module._cache;
 
   var dirname = path.dirname(filename);
@@ -411,14 +432,33 @@ Module.prototype._compile = function(content, filename) {
   return compiledWrapper.apply(self.exports, args);
 };
 
+
+function stripBOM(content) {
+  // Remove byte order marker. This catches EF BB BF (the UTF-8 BOM)
+  // because the buffer-to-string conversion in `fs.readFileSync()`
+  // translates it to FEFF, the UTF-16 BOM.
+  if (content.charCodeAt(0) === 0xFEFF) {
+    content = content.slice(1);
+  }
+  return content;
+}
+
+
 // Native extension for .js
 Module._extensions['.js'] = function(module, filename) {
   var content = NativeModule.require('fs').readFileSync(filename, 'utf8');
-  module._compile(content, filename);
+  module._compile(stripBOM(content), filename);
 };
 
 
-// Native extension for .node
+// Native extension for .json
+Module._extensions['.json'] = function(module, filename) {
+  var content = NativeModule.require('fs').readFileSync(filename, 'utf8');
+  module.exports = JSON.parse(stripBOM(content));
+};
+
+
+//Native extension for .node
 Module._extensions['.node'] = function(module, filename) {
   process.dlopen(filename, module.exports);
 };
@@ -439,10 +479,14 @@ Module._initPaths = function() {
   }
 
   if (process.env['NODE_PATH']) {
-    paths = process.env['NODE_PATH'].split(':').concat(paths);
+    var splitter = process.platform === 'win32' ? ';' : ':';
+    paths = process.env['NODE_PATH'].split(splitter).concat(paths);
   }
 
-  Module._paths = paths;
+  modulePaths = paths;
+
+  // clone as a read-only copy, for introspection.
+  Module.globalPaths = modulePaths.slice(0);
 };
 
 // bootstrap repl
