@@ -30,7 +30,7 @@ var isWindows = process.platform === 'win32';
 function normalizeArray(parts, allowAboveRoot) {
   // if the path tries to go above the root, `up` ends up > 0
   var up = 0;
-  for (var i = parts.length; i >= 0; i--) {
+  for (var i = parts.length - 1; i >= 0; i--) {
     var last = parts[i];
     if (last == '.') {
       parts.splice(i, 1);
@@ -55,15 +55,28 @@ function normalizeArray(parts, allowAboveRoot) {
 
 
 if (isWindows) {
-
-  // Regex to split a filename into [*, dir, basename, ext]
-  // windows version
-  var splitPathRe = /^(.+(?:[\\\/](?!$)|:)|[\\\/])?((?:.+?)?(\.[^.]*)?)$/;
-
   // Regex to split a windows path into three parts: [*, device, slash,
   // tail] windows-only
   var splitDeviceRe =
-      /^([a-zA-Z]:|[\\\/]{2}[^\\\/]+[\\\/][^\\\/]+)?([\\\/])?(.*?)$/;
+      /^([a-zA-Z]:|[\\\/]{2}[^\\\/]+[\\\/][^\\\/]+)?([\\\/])?([\s\S]*?)$/;
+
+  // Regex to split the tail part of the above into [*, dir, basename, ext]
+  var splitTailRe = /^([\s\S]+[\\\/](?!$)|[\\\/])?((?:\.{1,2}$|[\s\S]+?)?(\.[^.\/\\]*)?)$/;
+
+  // Function to split a filename into [root, dir, basename, ext]
+  // windows version
+  var splitPath = function(filename) {
+    // Separate device+slash from tail
+    var result = splitDeviceRe.exec(filename),
+        device = (result[1] || '') + (result[2] || ''),
+        tail = result[3] || '';
+    // Split the tail into dir, basename and extension
+    var result2 = splitTailRe.exec(tail),
+        dir = result2[1] || '',
+        basename = result2[2] || '',
+        ext = result2[3] || '';
+    return [device, dir, basename, ext];
+  };
 
   // path.resolve([from ...], to)
   // windows version
@@ -72,10 +85,19 @@ if (isWindows) {
         resolvedTail = '',
         resolvedAbsolute = false;
 
-    for (var i = arguments.length; i >= -1; i--) {
-      var path = (i >= 0)
-          ? arguments[i]
-          : process.cwd();
+    for (var i = arguments.length - 1; i >= -1; i--) {
+      var path;
+      if (i >= 0) {
+        path = arguments[i];
+      } else if (!resolvedDevice) {
+        path = process.cwd();
+      } else {
+        // Windows has the concept of drive-specific current working
+        // directories. If we've resolved a drive letter but not yet an
+        // absolute path, get cwd for that drive. We're sure the device is not
+        // an unc path at this points, because unc paths are always absolute.
+        path = process._cwdForDrive(resolvedDevice[0]);
+      }
 
       // Skip empty and invalid entries
       if (typeof path !== 'string' || !path) {
@@ -108,21 +130,6 @@ if (isWindows) {
       }
     }
 
-    if (!resolvedAbsolute && resolvedDevice) {
-      // If we still don't have an absolute path,
-      // prepend the current path for the device found.
-
-      // TODO
-      // Windows stores the current directories for 'other' drives
-      // as hidden environment variables like =C:=c:\windows (literally)
-      // var deviceCwd = os.getCwdForDrive(resolvedDevice);
-      var deviceCwd = '';
-
-      // If there is no cwd set for the drive, it is at root
-      resolvedTail = deviceCwd + '\\' + resolvedTail;
-      resolvedAbsolute = true;
-    }
-
     // Replace slashes (in UNC share name) by backslashes
     resolvedDevice = resolvedDevice.replace(/\//g, '\\');
 
@@ -186,12 +193,73 @@ if (isWindows) {
     return exports.normalize(joined);
   };
 
+  // path.relative(from, to)
+  // it will solve the relative path from 'from' to 'to', for instance:
+  // from = 'C:\\orandea\\test\\aaa'
+  // to = 'C:\\orandea\\impl\\bbb'
+  // The output of the function should be: '..\\..\\impl\\bbb'
+  // windows version
+  exports.relative = function(from, to) {
+    from = exports.resolve(from);
+    to = exports.resolve(to);
+
+    // windows is not case sensitive
+    var lowerFrom = from.toLowerCase();
+    var lowerTo = to.toLowerCase();
+
+    function trim(arr) {
+      var start = 0;
+      for (; start < arr.length; start++) {
+        if (arr[start] !== '') break;
+      }
+
+      var end = arr.length - 1;
+      for (; end >= 0; end--) {
+        if (arr[end] !== '') break;
+      }
+
+      if (start > end) return [];
+      return arr.slice(start, end - start + 1);
+    }
+
+    var toParts = trim(to.split('\\'));
+
+    var lowerFromParts = trim(lowerFrom.split('\\'));
+    var lowerToParts = trim(lowerTo.split('\\'));
+
+    var length = Math.min(lowerFromParts.length, lowerToParts.length);
+    var samePartsLength = length;
+    for (var i = 0; i < length; i++) {
+      if (lowerFromParts[i] !== lowerToParts[i]) {
+        samePartsLength = i;
+        break;
+      }
+    }
+
+    if (samePartsLength == 0) {
+      return to;
+    }
+
+    var outputParts = [];
+    for (var i = samePartsLength; i < lowerFromParts.length; i++) {
+      outputParts.push('..');
+    }
+
+    outputParts = outputParts.concat(toParts.slice(samePartsLength));
+
+    return outputParts.join('\\');
+  };
+
 
 } else /* posix */ {
 
-  // Regex to split a filename into [*, dir, basename, ext]
-  // posix version
-  var splitPathRe = /^(.+\/(?!$)|\/)?((?:.+?)?(\.[^.]*)?)$/;
+  // Split a filename into [root, dir, basename, ext], unix version
+  // 'root' is just a slash, or nothing.
+  var splitPathRe = /^(\/?)([\s\S]+\/(?!$)|\/)?((?:\.{1,2}$|[\s\S]+?)?(\.[^.\/]*)?)$/;
+  var splitPath = function(filename) {
+    var result = splitPathRe.exec(filename);
+    return [result[1] || '', result[2] || '', result[3] || '', result[4] || ''];
+  };
 
   // path.resolve([from ...], to)
   // posix version
@@ -199,10 +267,8 @@ if (isWindows) {
     var resolvedPath = '',
         resolvedAbsolute = false;
 
-    for (var i = arguments.length; i >= -1 && !resolvedAbsolute; i--) {
-      var path = (i >= 0)
-          ? arguments[i]
-          : process.cwd();
+    for (var i = arguments.length - 1; i >= -1 && !resolvedAbsolute; i--) {
+      var path = (i >= 0) ? arguments[i] : process.cwd();
 
       // Skip empty and invalid entries
       if (typeof path !== 'string' || !path) {
@@ -253,27 +319,75 @@ if (isWindows) {
       return p && typeof p === 'string';
     }).join('/'));
   };
+
+
+  // path.relative(from, to)
+  // posix version
+  exports.relative = function(from, to) {
+    from = exports.resolve(from).substr(1);
+    to = exports.resolve(to).substr(1);
+
+    function trim(arr) {
+      var start = 0;
+      for (; start < arr.length; start++) {
+        if (arr[start] !== '') break;
+      }
+
+      var end = arr.length - 1;
+      for (; end >= 0; end--) {
+        if (arr[end] !== '') break;
+      }
+
+      if (start > end) return [];
+      return arr.slice(start, end - start + 1);
+    }
+
+    var fromParts = trim(from.split('/'));
+    var toParts = trim(to.split('/'));
+
+    var length = Math.min(fromParts.length, toParts.length);
+    var samePartsLength = length;
+    for (var i = 0; i < length; i++) {
+      if (fromParts[i] !== toParts[i]) {
+        samePartsLength = i;
+        break;
+      }
+    }
+
+    var outputParts = [];
+    for (var i = samePartsLength; i < fromParts.length; i++) {
+      outputParts.push('..');
+    }
+
+    outputParts = outputParts.concat(toParts.slice(samePartsLength));
+
+    return outputParts.join('/');
+  };
+
 }
 
 
 exports.dirname = function(path) {
-  var dir = splitPathRe.exec(path)[1] || '';
-  if (!dir) {
-    // No dirname
+  var result = splitPath(path),
+      root = result[0],
+      dir = result[1];
+
+  if (!root && !dir) {
+    // No dirname whatsoever
     return '.';
-  } else if (dir.length === 1 ||
-      (isWindows && dir.length <= 3 && dir.charAt(1) === ':')) {
-    // It is just a slash or a drive letter with a slash
-    return dir;
-  } else {
-    // It is a full dirname, strip trailing slash
-    return dir.substring(0, dir.length - 1);
   }
+
+  if (dir) {
+    // It has a dirname, strip trailing slash
+    dir = dir.substring(0, dir.length - 1);
+  }
+
+  return root + dir;
 };
 
 
 exports.basename = function(path, ext) {
-  var f = splitPathRe.exec(path)[2] || '';
+  var f = splitPath(path)[2];
   // TODO: make this comparison case-insensitive on windows?
   if (ext && f.substr(-1 * ext.length) === ext) {
     f = f.substr(0, f.length - ext.length);
@@ -283,7 +397,7 @@ exports.basename = function(path, ext) {
 
 
 exports.extname = function(path) {
-  return splitPathRe.exec(path)[3] || '';
+  return splitPath(path)[3];
 };
 
 
@@ -302,3 +416,24 @@ exports.existsSync = function(path) {
     return false;
   }
 };
+
+
+exports._makeLong = isWindows ? 
+  function(path) {
+    var resolvedPath = exports.resolve(path);
+
+    if (resolvedPath.match(/^[a-zA-Z]\:\\/)) {
+      // path is local filesystem path, which needs to be converted
+      // to long UNC path.
+      return '\\\\?\\' + resolvedPath;
+    } else if (resolvedPath.match(/^\\\\[^?.]/)) {
+      // path is network UNC path, which needs to be converted
+      // to long UNC path.
+      return '\\\\?\\UNC\\' + resolvedPath.substring(2);
+    }
+
+    return path;
+  } :
+  function(path) {
+    return path;
+  };
