@@ -42,6 +42,31 @@ try {
   throw new Error('node.js not compiled with openssl crypto support.');
 }
 
+// Convert protocols array into valid OpenSSL protocols list
+// ("\x06spdy/2\x08http/1.1\x08http/1.0")
+function convertNPNProtocols(NPNProtocols, out) {
+  // If NPNProtocols is Array - translate it into buffer
+  if (Array.isArray(NPNProtocols)) {
+    var buff = new Buffer(NPNProtocols.reduce(function(p, c) {
+      return p + 1 + Buffer.byteLength(c);
+    }, 0));
+
+    NPNProtocols.reduce(function(offset, c) {
+      var clen = Buffer.byteLength(c);
+      buff[offset] = clen;
+      buff.write(c, offset + 1);
+
+      return offset + 1 + clen;
+    }, 0);
+
+    NPNProtocols = buff;
+  }
+
+  // If it's already a Buffer - store it
+  if (Buffer.isBuffer(NPNProtocols)) {
+    out.NPNProtocols = NPNProtocols;
+  }
+}
 
 // Base class of both CleartextStream and EncryptedStream
 function CryptoStream(pair) {
@@ -52,6 +77,7 @@ function CryptoStream(pair) {
   this.readable = this.writable = true;
 
   this._paused = false;
+  this._needDrain = false;
   this._pending = [];
   this._pendingCallbacks = [];
   this._pendingBytes = 0;
@@ -86,7 +112,7 @@ CryptoStream.prototype.write = function(data /* , encoding, cb */) {
     data = new Buffer(data, encoding);
   }
 
-  debug('clearIn data');
+  debug((this === this.pair.cleartext ? 'clear' : 'encrypted') + 'In data');
 
   this._pending.push(data);
   this._pendingCallbacks.push(cb);
@@ -95,7 +121,26 @@ CryptoStream.prototype.write = function(data /* , encoding, cb */) {
   this.pair._writeCalled = true;
   this.pair.cycle();
 
-  return this._pendingBytes < 128 * 1024;
+  // In the following cases, write() should return a false,
+  // then this stream should eventually emit 'drain' event.
+  //
+  // 1. There are pending data more than 128k bytes.
+  // 2. A forward stream shown below is paused.
+  //    A) EncryptedStream for CleartextStream.write().
+  //    B) CleartextStream for EncryptedStream.write().
+  //
+  if (!this._needDrain) {
+    if (this._pendingBytes >= 128 * 1024) {
+      this._needDrain = true;
+    } else {
+      if (this === this.pair.cleartext) {
+        this._needDrain = this.pair.encrypted._paused;
+      } else {
+        this._needDrain = this.pair.cleartext._paused;
+      }
+    }
+  }
+  return !this._needDrain;
 };
 
 
@@ -128,12 +173,12 @@ CryptoStream.prototype.setEncoding = function(encoding) {
 };
 
 
-// EG '/C=US/ST=CA/L=SF/O=Joyent/OU=Node.js/CN=ca1/emailAddress=ry@clouds.org'
+// Example:
+// C=US\nST=CA\nL=SF\nO=Joyent\nOU=Node.js\nCN=ca1\nemailAddress=ry@clouds.org
 function parseCertString(s) {
   var out = {};
-  var parts = s.split('/');
-  // Note: can always skip the first one.
-  for (var i = 1; i < parts.length; i++) {
+  var parts = s.split('\n');
+  for (var i = 0, len = parts.length; i < len; i++) {
     var sepIndex = parts[i].indexOf('=');
     if (sepIndex > 0) {
       var key = parts[i].slice(0, sepIndex);
@@ -159,6 +204,21 @@ CryptoStream.prototype.getPeerCertificate = function() {
   return null;
 };
 
+CryptoStream.prototype.getSession = function() {
+  if (this.pair.ssl) {
+    return this.pair.ssl.getSession();
+  }
+
+  return null;
+};
+
+CryptoStream.prototype.isSessionReused = function() {
+  if (this.pair.ssl) {
+    return this.pair.ssl.isSessionReused();
+  }
+
+  return null;
+};
 
 CryptoStream.prototype.getCipher = function(err) {
   if (this.pair.ssl) {
@@ -211,7 +271,11 @@ CryptoStream.prototype._done = function() {
       this.pair.encrypted._doneFlag &&
       !this.pair._doneFlag) {
     // If both streams are done:
-    this.pair.destroy();
+    if (!this.pair._secureEstablished) {
+      this.pair.error();
+    } else {
+      this.pair.destroy();
+    }
   }
 };
 
@@ -351,7 +415,7 @@ CryptoStream.prototype._pull = function() {
         this.pair.encrypted._destroyAfterPush = true;
       }
       this.pair.cycle();
-      this._done()
+      this._done();
       return;
     }
 
@@ -380,11 +444,31 @@ CryptoStream.prototype._pull = function() {
     assert(rv === tmp.length);
   }
 
-  // If we've cleared all of incoming encrypted data, emit drain.
-  if (havePending && this._pending.length === 0) {
-    debug('drain');
-    this.emit('drain');
-    if (this.__destroyOnDrain) this.end();
+  // If pending data has cleared, 'drain' event should be emitted
+  // after write() returns a false.
+  // Except when a forward stream shown below is paused.
+  //   A) EncryptedStream for CleartextStream._pull().
+  //   B) CleartextStream for EncryptedStream._pull().
+  //
+  if (this._needDrain && this._pending.length === 0) {
+    var paused;
+    if (this === this.pair.cleartext) {
+      paused = this.pair.encrypted._paused;
+    } else {
+      paused = this.pair.cleartext._paused;
+    }
+    if (!paused) {
+      debug('drain ' + (this === this.pair.cleartext ? 'clear' : 'encrypted'));
+      var self = this;
+      process.nextTick(function() {
+        if (typeof self.ondrain === 'function') {
+          self.ondrain();
+        }
+        self.emit('drain');
+      });
+      this._needDrain = false;
+      if (this.__destroyOnDrain) this.end();
+    }
   }
 };
 
@@ -416,6 +500,19 @@ CleartextStream.prototype._pusher = function(pool, offset, length) {
   return this.pair.ssl.clearOut(pool, offset, length);
 };
 
+CleartextStream.prototype.address = function() {
+  return this.socket && this.socket.address();
+};
+
+CleartextStream.prototype.__defineGetter__('remoteAddress', function() {
+  return this.socket && this.socket.remoteAddress;
+});
+
+
+CleartextStream.prototype.__defineGetter__('remotePort', function() {
+  return this.socket && this.socket.remotePort;
+});
+
 
 function EncryptedStream(pair) {
   CryptoStream.call(this, pair);
@@ -449,16 +546,20 @@ EncryptedStream.prototype._pusher = function(pool, offset, length) {
  * Provides a pair of streams to do encrypted communication.
  */
 
-function SecurePair(credentials, isServer, requestCert, rejectUnauthorized) {
+function SecurePair(credentials, isServer, requestCert, rejectUnauthorized,
+                    options) {
   if (!(this instanceof SecurePair)) {
     return new SecurePair(credentials,
                           isServer,
                           requestCert,
-                          rejectUnauthorized);
+                          rejectUnauthorized,
+                          options);
   }
 
   var self = this;
 
+  options || (options = {});
+
   events.EventEmitter.call(this);
 
   this._secureEstablished = false;
@@ -479,15 +580,25 @@ function SecurePair(credentials, isServer, requestCert, rejectUnauthorized) {
     requestCert = true;
   }
 
-  this._secureEstablished = false;
   this._rejectUnauthorized = rejectUnauthorized ? true : false;
   this._requestCert = requestCert ? true : false;
 
   this.ssl = new Connection(this.credentials.context,
-                             this._isServer ? true : false,
-                             this._requestCert,
-                             this._rejectUnauthorized);
+      this._isServer ? true : false,
+      this._isServer ? this._requestCert : options.servername,
+      this._rejectUnauthorized);
 
+  if (process.features.tls_sni) {
+    if (this._isServer && options.SNICallback) {
+      this.ssl.setSNICallback(options.SNICallback);
+    }
+    this.servername = null;
+  }
+
+  if (process.features.tls_npn && options.NPNProtocols) {
+    this.ssl.setNPNProtocols(options.NPNProtocols);
+    this.npnProtocol = null;
+  }
 
   /* Acts as a r/w stream to the cleartext side of the stream. */
   this.cleartext = new CleartextStream(this);
@@ -559,28 +670,28 @@ SecurePair.prototype.cycle = function(depth) {
 
   if (!this.cycleEncryptedPullLock) {
     this.cycleEncryptedPullLock = true;
-    debug("encrypted._pull");
+    debug('encrypted._pull');
     this.encrypted._pull();
     this.cycleEncryptedPullLock = false;
   }
 
   if (!this.cycleCleartextPullLock) {
     this.cycleCleartextPullLock = true;
-    debug("cleartext._pull");
+    debug('cleartext._pull');
     this.cleartext._pull();
     this.cycleCleartextPullLock = false;
   }
 
   if (!this.cycleCleartextPushLock) {
     this.cycleCleartextPushLock = true;
-    debug("cleartext._push");
+    debug('cleartext._push');
     this.cleartext._push();
     this.cycleCleartextPushLock = false;
   }
 
   if (!this.cycleEncryptedPushLock) {
     this.cycleEncryptedPushLock = true;
-    debug("encrypted._push");
+    debug('encrypted._push');
     this.encrypted._push();
     this.cycleEncryptedPushLock = false;
   }
@@ -596,6 +707,14 @@ SecurePair.prototype.cycle = function(depth) {
 
 SecurePair.prototype.maybeInitFinished = function() {
   if (this.ssl && !this._secureEstablished && this.ssl.isInitFinished()) {
+    if (process.features.tls_npn) {
+      this.npnProtocol = this.ssl.getNegotiatedProtocol();
+    }
+
+    if (process.features.tls_sni) {
+      this.servername = this.ssl.getServername();
+    }
+
     this._secureEstablished = true;
     debug('secure established');
     this.emit('secure');
@@ -616,6 +735,7 @@ SecurePair.prototype.destroy = function() {
     self.cleartext.writable = self.cleartext.readable = false;
 
     process.nextTick(function() {
+      self.cleartext.emit('end');
       self.encrypted.emit('close');
       self.cleartext.emit('close');
     });
@@ -625,7 +745,13 @@ SecurePair.prototype.destroy = function() {
 
 SecurePair.prototype.error = function() {
   if (!this._secureEstablished) {
+    var error = this.ssl.error;
+    if (!error) {
+      error = new Error('socket hang up');
+      error.code = 'ECONNRESET';
+    }
     this.destroy();
+    this.emit('error', error);
   } else {
     var err = this.ssl.error;
     this.ssl.error = null;
@@ -735,15 +861,16 @@ function Server(/* [options], listener */) {
 
   var sharedCreds = crypto.createCredentials({
     key: self.key,
+    passphrase: self.passphrase,
     cert: self.cert,
     ca: self.ca,
-    ciphers: self.ciphers,
+    ciphers: self.ciphers || 'RC4-SHA:AES128-SHA:AES256-SHA',
     secureProtocol: self.secureProtocol,
-    crl: self.crl
+    secureOptions: self.secureOptions,
+    crl: self.crl,
+    sessionIdContext: self.sessionIdContext
   });
 
-  sharedCreds.context.setCiphers('RC4-SHA:AES128-SHA:AES256-SHA');
-
   // constructor call
   net.Server.call(this, function(socket) {
     var creds = crypto.createCredentials(null, sharedCreds.context);
@@ -751,13 +878,20 @@ function Server(/* [options], listener */) {
     var pair = new SecurePair(creds,
                               true,
                               self.requestCert,
-                              self.rejectUnauthorized);
+                              self.rejectUnauthorized,
+                              {
+                                NPNProtocols: self.NPNProtocols,
+                                SNICallback: self.SNICallback
+                              });
 
     var cleartext = pipe(pair, socket);
     cleartext._controlReleased = false;
 
     pair.on('secure', function() {
       pair.cleartext.authorized = false;
+      pair.cleartext.npnProtocol = pair.npnProtocol;
+      pair.cleartext.servername = pair.servername;
+
       if (!self.requestCert) {
         cleartext._controlReleased = true;
         self.emit('secureConnection', pair.cleartext, pair.encrypted);
@@ -780,14 +914,14 @@ function Server(/* [options], listener */) {
         }
       }
     });
+    pair.on('error', function(err) {
+      self.emit('clientError', err);
+    });
   });
 
   if (listener) {
     this.on('secureConnection', listener);
   }
-
-  // Handle option defaults:
-  this.setOptions(options);
 }
 
 util.inherits(Server, net.Server);
@@ -811,10 +945,54 @@ Server.prototype.setOptions = function(options) {
   }
 
   if (options.key) this.key = options.key;
+  if (options.passphrase) this.passphrase = options.passphrase;
   if (options.cert) this.cert = options.cert;
   if (options.ca) this.ca = options.ca;
   if (options.secureProtocol) this.secureProtocol = options.secureProtocol;
   if (options.crl) this.crl = options.crl;
+  if (options.ciphers) this.ciphers = options.ciphers;
+  if (options.secureProtocol) this.secureProtocol = options.secureProtocol;
+  if (options.secureOptions) this.secureOptions = options.secureOptions;
+  if (options.NPNProtocols) convertNPNProtocols(options.NPNProtocols, this);
+  if (options.SNICallback) {
+    this.SNICallback = options.SNICallback;
+  } else {
+    this.SNICallback = this.SNICallback.bind(this);
+  }
+  if (options.sessionIdContext) {
+    this.sessionIdContext = options.sessionIdContext;
+  } else if (this.requestCert) {
+    this.sessionIdContext = crypto.createHash('md5')
+                                  .update(process.argv.join(' '))
+                                  .digest('hex');
+  }
+};
+
+// SNI Contexts High-Level API
+Server.prototype._contexts = [];
+Server.prototype.addContext = function(servername, credentials) {
+  if (!servername) {
+    throw 'Servername is required parameter for Server.addContext';
+  }
+
+  var re = new RegExp('^' +
+                      servername.replace(/([\.^$+?\-\\[\]{}])/g, '\\$1')
+                                .replace(/\*/g, '.*') +
+                      '$');
+  this._contexts.push([re, crypto.createCredentials(credentials).context]);
+};
+
+Server.prototype.SNICallback = function(servername) {
+  var ctx;
+
+  this._contexts.some(function(elem) {
+    if (servername.match(elem[0]) !== null) {
+      ctx = elem[1];
+      return true;
+    }
+  });
+
+  return ctx;
 };
 
 
@@ -855,17 +1033,30 @@ exports.connect = function(port /* host, options, cb */) {
   var socket = new net.Stream();
 
   var sslcontext = crypto.createCredentials(options);
-  //sslcontext.context.setCiphers('RC4-SHA:AES128-SHA:AES256-SHA');
 
-  var pair = new SecurePair(sslcontext, false);
+  convertNPNProtocols(options.NPNProtocols, this);
+  var pair = new SecurePair(sslcontext, false, true, false,
+                            {
+                              NPNProtocols: this.NPNProtocols,
+                              servername: options.servername || host
+                            });
+
+  if (options.session) {
+    pair.ssl.setSession(options.session);
+  }
 
   var cleartext = pipe(pair, socket);
+  if (cb) {
+    cleartext.on('secureConnect', cb);
+  }
 
   socket.connect(port, host);
 
   pair.on('secure', function() {
     var verifyError = pair.ssl.verifyError();
 
+    cleartext.npnProtocol = pair.npnProtocol;
+
     if (verifyError) {
       cleartext.authorized = false;
       cleartext.authorizationError = verifyError;
@@ -873,7 +1064,10 @@ exports.connect = function(port /* host, options, cb */) {
       cleartext.authorized = true;
     }
 
-    if (cb) cb();
+    cleartext.emit('secureConnect');
+  });
+  pair.on('error', function(err) {
+    cleartext.emit('error', err);
   });
 
   cleartext._controlReleased = true;
@@ -899,7 +1093,7 @@ function pipe(pair, socket) {
 
   function onclose() {
     socket.removeListener('error', onerror);
-    socket.removeListener('close', onclose);
+    socket.removeListener('end', onclose);
     socket.removeListener('timeout', ontimeout);
   }
 
