@@ -22,14 +22,15 @@
 // Can't test this when 'make test' doesn't assign a tty to the stdout.
 var common = require('../common');
 var assert = require('assert');
-var tty = require('tty');
 
-var closed = false;
-process.stdout.on('close', function() {
-  closed = true;
-});
-process.on('exit', function() {
-  assert.ok(closed);
-});
+var exceptionCaught = false;
 
-process.stdout.end();
+try {
+  process.stdout.end();
+} catch (e) {
+  exceptionCaught = true;
+  assert.ok(common.isError(e));
+  assert.equal('process.stdout cannot be closed', e.message);
+}
+
+assert.ok(exceptionCaught);
